/****************************************************************************
*    This file is part of CEDEC_GO.                                         *
*    Copyright (c) 2004-2011 Takenori Kubo, Youichiro Miyake, CESA          *
*                                                                           *
*    History:                                                               *
*    01/07/2011 Tweaked for CEDEC CHALLANGE 2011 by CESA                    *
*                                                                           *
*    CEDEC_GO is free software: you can redistribute it and/or modify       *
*    it under the terms of the GNU General Public License as published by   *
*    the Free Software Foundation, either version 3 of the License, or      *
*    any later version.                                                     *
*                                                                           *
*    CEDEC_GO is distributed in the hope that it will be useful,            *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of         *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
*    GNU General Public License for more details.                           *
*                                                                           *
*    You should have received a copy of the GNU General Public License      *
*    along with CEDEC_GO.  If not, see <http://www.gnu.org/licenses/>.      *
*****************************************************************************/

// GTP_IGO_02.cpp : コンソール アプリケーションのエントリ ポイントを定義します。
// GTP_IGO_02.cpp : This file defines Entry Point for Console Application.

#include "stdafx.h" // ここにあったインクルード・ライブラリは "stdafx.h"に移動しました。 // Include files are moved to "stdafx.h" // Youichiro Miyake

#include "time.h"
#include "board.h"

class sample_engine :public engine {
  int size;
  float komi;
  struct move {
    int x;
    int y;
    COLOR color;    
    move(int _x, int _y, COLOR _color)
      :x(_x), y(_y), color(_color)
    {}
    ~move(){}
  };
  std::vector<move> moves;
  COLOR board[MAX_BOARD][MAX_BOARD];
  COLOR status[MAX_BOARD][MAX_BOARD];

  cboard cboard;

 public:
  sample_engine()
    :size(19), komi(0), moves()
  {
    reset(size, komi);
  }
  virtual ~sample_engine() {}

  virtual void reset(int size, float komi) {
    this->komi = komi;
    this->size = size;
    for (int i=0; i < MAX_BOARD; i++) {
      for (int j=0; j < MAX_BOARD; j++) {
		board[i][j] = EMPTY;
      }
    }
	moves.clear();

	cboard.init( size );
  }

  virtual std::string name() {
    return ENTRY_NAME; // プログラム名をヘッダーファイルで定義してください。 Youichiro Miyake // Define your program name in Header File.
  }
  virtual std::string version() {
    return CEDEC_VERSION; // ヘッダーファイルで定義してください。 Youichiro Miyake // Define this in Header File.
  }

  //////////////////////////////////////////////////////////////////////////////////////////////////
  // 注：この関数が手を決定する思考関数です。この関数を書き換えてください。
  // デフォルトでは左上から石のない場所に順番に並べて行くプログラムになっています。Youichiro Miyake
  //
  // Attention : This function decides your next move. Modify this function.
  // In default setting, this function returns the move in order from left-top to right-down for empty place. 
  //
  //---------------------------------------------------------------//

  virtual void genmove(int& i, int& j, COLOR color){
	// short move = cboard.think( board, size, color );
	short move = cboard.think( size, color );

	if ( move == 0 ) {
	  // PASS
	  i = -1;
	  j = -1;
	} else {
	  i = move / ( size + 1 ) - 1;
	  j = move % ( size + 1 ) - 1;
	}

	return;
  }

 //---------------------------------------------------------------//

  virtual void playmove(int i, int j, COLOR color){
	moves.push_back(move(i, j, color));
	board[i][j] = color;

	cboard.put( ( color == BLACK ? BALLADE_ME : BALLADE_YOU ),
				( j + 1 ) + ( i + 1 ) * ( cboard.mSize + 1 ) );

	return;
  }

  virtual void undo(){
    move m = moves.back();
    board[m.x][m.y] = EMPTY;
    moves.pop_back();
  }

  virtual bool legal_move(int i, int j, COLOR color){
    return board[i][j] == EMPTY;
  }

  virtual void compute_final_status() {
    for (int i = 0; i < size; i++) {
      for (int j = 0; j < size; j++) {
	if (board[i][j] == BLACK) {
	  status[i][j] = BLACK_TERRITORY;
	} else if (board[i][j] == WHITE) {
	  status[i][j] = WHITE_TERRITORY;
	} else {
	  status[i][j] = EMPTY;
	}
      }
    }
  }
  
  virtual COLOR get_board(int i, int j) {
    return board[i][j];
  }
  virtual COLOR get_final_status(int i, int j) {
    return status[i][j];
  }

  virtual void time_settings(int main_time,
			     int byo_yomi_time,
			     int byo_yomi_stones) {
  }
  virtual void time_left(COLOR color, int time, int stones) {
  }


};

/* この部分は Windows コンソール向けに以下のように書き換えます。Youichiro Miyake
   This part is modified for Windows Console Program //

int
main(void)
{
  gtp g(new sample_engine);
  g.main_loop();
  return 0;
}
*/

int _tmain(int argc, _TCHAR* argv[])
{
    
    gtp g(new sample_engine);
    g.main_loop();

	return 0;
}
