#include <windows.h>
#include <mmsystem.h>
#include <time.h>
#include "opening.h"

copening opening;

bool cboard::init( int iSize ) {
  srand( ( unsigned int ) time( 0 ) );

  for ( int i = 0; i < BALLADE_MAX_BOARD; i++ )
	mBoard[ i ] = BALLADE_OUTSIDE;

  ( ( TotalBoard* ) &mBoard )->capturedStones[ 0 ] = 0;
  ( ( TotalBoard* ) &mBoard )->capturedStones[ 1 ] = 0;
  ( ( TotalBoard* ) &mBoard )->ko    = 0;
  ( ( TotalBoard* ) &mBoard )->pass  = 0;

  mSize = iSize;

  for ( int j = 1; j <= iSize; j++ )
	for ( int i = 1; i <= iSize; i++ )
	  mBoard[ i + j * ( iSize + 1 ) ] = BALLADE_EMPTY;

  opening.init();

  return true;
}

bool cboard::invRotMir( int x, int y, int rot, int mir, int *x1, int *y1 ) {
  if ( mir )
	x = mSize + 1 - x;

  rot = ( 4 - rot ) % 4;

  for ( int r = 0; r < rot; r++ ) {
	int xx = y;
	int yy = mSize + 1 - x;
	x = xx;
	y = yy;
  }

  *x1 = x;
  *y1 = y;

  return true;
}

bool cboard::rotate( unsigned char *board0,
					 unsigned char *board1 ) {
  memcpy( board1, board0, BALLADE_MAX_BOARD );
  for ( int j = 1; j <= mSize; j++ )
	for ( int i = 1; i <= mSize; i++ )
	  board1[ j + ( mSize + 1 - i ) * ( mSize + 1 ) ]
		= board0[ i + j * ( mSize + 1 ) ];
  return true;
}

bool cboard::mirror( unsigned char *board0,
					 unsigned char *board1 ) {
  memcpy( board1, board0, BALLADE_MAX_BOARD );
  for ( int j = 1; j <= mSize; j++ )
	for ( int i = 1; i <= mSize; i++ )
	  board1[ ( mSize + 1 - i ) + j * ( mSize + 1 ) ]
		= board0[ i + j * ( mSize + 1 ) ];
  return true;
}

bool cboard::compress( void ) {
  unsigned char boardr[ BALLADE_MAX_BOARD ];
  unsigned char boardm[ BALLADE_MAX_BOARD ];

  memset( mCompressed, 0xff, COMPRESSED_BOARD_SIZE );

  for ( int rot = 0; rot < 4; rot++ ) {
	if ( rot == 0 )
	  memcpy( boardr, mBoard, BALLADE_MAX_BOARD );
	else {
	  rotate( boardr, boardm );
	  memcpy( boardr, boardm, BALLADE_MAX_BOARD );
	}

	for ( int mir = 0; mir < 2; mir++ ) {
	  if ( mir == 0 )
		memcpy( boardm, boardr, BALLADE_MAX_BOARD );
	  else
		mirror( boardr, boardm );

	  unsigned char cmp[ COMPRESSED_BOARD_SIZE ];
	  compressCore( boardm, cmp );

	  if ( memcmp( mCompressed, cmp, COMPRESSED_BOARD_SIZE ) > 0 ) {
		memcpy( mCompressed, cmp, COMPRESSED_BOARD_SIZE );
		mCompressRotate	=	rot;
		mCompressMirror	=	mir;
	  }
	}
  }

  return true;
}

bool cboard::compressCore( unsigned char *board0,
						   unsigned char *board1 ) {
  int p = 0;
  unsigned char *p1 = board1;

  for ( int j = 1; j <= mSize; j++ )
	for ( int i = 1; i <= mSize; i++ ) {
	  if ( p == 0 )
		*p1 = 0;

	  *p1 = *p1 * 3 + board0[ i + j * ( mSize + 1 ) ];

	  p = ++p % 5;
	  if ( p == 0 )
		p1++;
	}
  if ( p )
	for ( int i = p; i < 5; i++ )
	  *p1 *= 3;

  return true;
}

int cboard::put( int stone,
				 int pos ) {

  if ( pos != 0 && ( ( TotalBoard* ) mBoard ) -> ko == pos )
	return ERROR_KO;
  ( ( TotalBoard* ) mBoard ) -> ko = 0;

  pCapturedStone = 0; 

  if ( pos == BALLADE_PASS )
	return 0;

  if ( mBoard[ pos ] != (BYTE)NULL )
	return ERROR_KOMA;

  int count;

  if ( ( count = getAll( mBoard, pos, stone ) ) != 0 ) {
	if ( count == 1 ) {
	  int n = 0;
	  uchar you = BALLADE_ME + BALLADE_YOU - stone;
	  uchar st;
	  st = mBoard[ pos         + 1 ];
	  if ( st == you || st == BALLADE_OUTSIDE ) n++;
	  st = mBoard[ pos         - 1 ];
	  if ( st == you || st == BALLADE_OUTSIDE ) n++;
	  st = mBoard[ pos + mSize + 1 ];
	  if ( st == you || st == BALLADE_OUTSIDE ) n++;
	  st = mBoard[ pos - mSize - 1 ];
	  if ( st == you || st == BALLADE_OUTSIDE ) n++;
	  if ( n == 3 )
		( ( TotalBoard* ) mBoard ) -> ko = mCapturedStone[ pCapturedStone - 1 ];
	}
  } else if ( suicide( mBoard, pos, stone ) )
	return ERROR_SUICIDE;

  mBoard[ pos ] = stone;

  ( ( TotalBoard* ) &mBoard ) -> capturedStones[ stone - 1 ] += count;

  return 0;
}

short cboard::think( /*COLOR board[ MAX_BOARD ][ MAX_BOARD ],*/
					int iSize,
					COLOR color ) {
  /*
  memset( mBoard, BALLADE_OUTSIDE, BALLADE_MAX_BOARD );
  for ( int j = 0; j < iSize; j++ )
	for ( int i = 0; i < iSize; i++ )
	  if ( board[ j ][ i ] == color )
		mBoard[ ( i + 1 ) + ( j + 1 ) * ( iSize + 1 ) ] = BALLADE_ME;
	  else if ( board[ j ][ i ] + color == WHITE + BLACK )
		mBoard[ ( i + 1 ) + ( j + 1 ) * ( iSize + 1 ) ] = BALLADE_YOU;
	  else
	  mBoard[ ( i + 1 ) + ( j + 1 ) * ( iSize + 1 ) ] = BALLADE_EMPTY;
  */

  mNeighbour[ 0 ] = - iSize - 1;
  mNeighbour[ 1 ] =         - 1;
  mNeighbour[ 2 ] =           1;
  mNeighbour[ 3 ] =   iSize + 1;

  mSize = iSize;

  MoveList bestMove;

  //---------------- opening
  // unsigned char compressed[ COMPRESSED_BOARD_SIZE ];
  // unsigned char rotate, mirror;
  // compress( compressed, &rotate, &mirror );
  compress();
  int popening = opening.find( mCompressed );
  if ( popening >= 0 ) {
	int best = opening.m_book[ popening ].move;
	int bestX = best % ( iSize + 1 );
	int bestY = best / ( iSize + 1 );
	int x1, y1;
	int pos;
	if ( best != 0 ) {
	  invRotMir( bestX,
				 bestY,
				 mCompressRotate,
				 mCompressMirror,
				 &x1,	&y1 );
	  pos = x1 + y1 * ( iSize + 1 );
	} else
	  pos = 0;
	bestMove.move = pos;
  } else
	bestMove = mc();

  return bestMove.move;
}

short
cboard::liberty( uchar *board,
				 short pos ) {
  BYTE liberty[ BALLADE_MAX_BOARD ];
  memset( liberty, 0, BALLADE_MAX_BOARD );
  short stack[ BALLADE_MAX_BOARD ], count = 0, p = 1;
  stack[ 0 ] = pos;
  unsigned char me = board[ pos ];
  liberty[ pos ] = 1;
  do {
    short pos0 = stack[ --p ];
    for ( int i = 0; i < 4; i++ ) {
      short pos1 = pos0 + mNeighbour[ i ];
      if ( board[ pos1 ] == BALLADE_EMPTY
	   && liberty[ pos1 ] == 0 ) {
		count++;
		liberty[ pos1 ] = 1;
      } else if ( board[ pos1 ] == me
				  && liberty[ pos1 ] == 0 ) {
		stack[ p++ ] = pos1;
		liberty[ pos1 ] = 1;
	  }
	}
  } while ( p > 0 );
  return count;
}

bool
cboard::canPut( int    me,
				int    pos,
				uchar* board ) {

  unsigned char you = BALLADE_ME + BALLADE_YOU - me;

  if ( pos == BALLADE_PASS )
    return true;

  //if ( ( ( TotalBoard* ) board ) -> ko != 0 )
  /*if ( ( ( TotalBoard* ) board ) -> ko == pos )
    return false;*/

  if ( board[ pos ] != BALLADE_EMPTY )
    return false;

  for ( int i = 0; i < 4; i++ ) {
	int pos1 = pos + mNeighbour[ i ];
    if ( board[ pos1 ] == BALLADE_EMPTY )
      return true;
    else if ( board[ pos1 ] == you )
	  if ( liberty( board, pos1 ) == 1 )
	return true;
  }

  board[ pos ] = me;
  if ( liberty( board, pos ) == 0 ) {
	board[ pos ] = BALLADE_EMPTY;
	return false;
  }
  board[ pos ] = BALLADE_EMPTY;

  return true;
}

bool
cboard::canPut1( int    me,
				 int    pos,
				 uchar* board ) {

  if ( ! canPut( me, pos, board ) )
	return false;

  /*
  for ( int j = -1; j <= 1; j++ )
    for ( int i = -1; i <= 1; i++ )
      if ( i || j )
	if ( board[ pos + i + j * ( mWidth + 1 ) ] != me &&
	     board[ pos + i + j * ( mWidth + 1 ) ] != OUTSIDE )
	  return true;
  */

  if ( board[ pos         - 1 ] != me &&
       board[ pos         - 1 ] != BALLADE_OUTSIDE )
    return true;
  if ( board[ pos         + 1 ] != me &&
       board[ pos         + 1 ] != BALLADE_OUTSIDE )
    return true;
  if ( board[ pos + mSize + 1 ] != me &&
       board[ pos + mSize + 1 ] != BALLADE_OUTSIDE )
	return true;
  if ( board[ pos - mSize - 1 ] != me &&
       board[ pos - mSize - 1 ] != BALLADE_OUTSIDE )
    return true;

  board[ pos ] = me;
  if ( liberty( board, pos ) > 1 ) {
	board[ pos ] = BALLADE_EMPTY;
    return true;
  }
  board[ pos ] = BALLADE_EMPTY;

  return false;
}

short cboard::liberty1( BYTE *board, short pos ) {
  short stack[ BALLADE_MAX_BOARD ];
  short liberty = -1;
  int count = 0, p0 = 0, p1 = 1;
  stack[ 0 ] = pos;
  BYTE me = board[ pos ];

  do {
    short pos0 = stack[ p0 ];
    for ( int i = 0; i < 4; i++ ) {
      short pos1 = pos0 + mNeighbour[ i ];
      if ( board[ pos1 ] == (BYTE)NULL && pos1 != liberty ) {
	if ( ++count > 1 )
	  return 2;
	liberty = pos1;
      } else if ( board[ pos1 ] == me ) {
	bool flag = false;
	for ( int i = 0; i < p1; i++ )
	  if ( pos1 == stack[ i ] ) {
	    flag = true;
	    break;
	  }
	if ( flag == false )
	  stack[ p1++ ] = pos1;
      }
    }
  } while ( ++p0 < p1 );

  return count;
}

bool cboard::suicide( BYTE *board, short pos, int stone ) {
  for ( int i = 0; i < 4; i++ )
    if ( board[ pos + mNeighbour[ i ] ] == 0 )
	  return false;
  {
	for ( int i = 0; i < 4; i++ ) {
	  short pos1 = pos + mNeighbour[ i ];
	  if ( board[ pos1 ] == stone && liberty1( board, pos1 ) > 1 )
		return false;
	  /*else if ( board[ pos1 ] == you && liberty1( board, pos1 ) == 1 )
		return false;*/
	}
  }
  return true;
}

int cboard::put2( uchar *board,
				  int stone,
				  int pos ) {

  if ( pos == BALLADE_PASS ) // pass
    return 0;

  /*  if ( ( ( TotalBoard* ) board ) -> ko == pos )
	  return ERROR_KO;
	  ( ( TotalBoard* ) board ) -> ko = 0;*/

  pCapturedStone = 0;

  if ( board[ pos ] != (BYTE)NULL )
    return ERROR_KOMA;

  int count;
  if ( ( count = getAll( board, pos, stone ) ) != 0 ) {
    if ( count == 1 ) {
      int n = 0;
      uchar you = BALLADE_ME + BALLADE_YOU - stone;
      uchar st;
      st = board[ pos         + 1 ]; if ( st == you || st == BALLADE_OUTSIDE ) n++;
      st = board[ pos         - 1 ]; if ( st == you || st == BALLADE_OUTSIDE ) n++;
      st = board[ pos + mSize + 1 ]; if ( st == you || st == BALLADE_OUTSIDE ) n++;
      st = board[ pos - mSize - 1 ]; if ( st == you || st == BALLADE_OUTSIDE ) n++;
	  /*
      if ( n == 3 ) {
		( ( TotalBoard* ) board ) -> ko = mCapturedStone[ pCapturedStone - 1 ];
      }
	  */
    }
  } else if ( suicide( board, pos, stone ) )
    return ERROR_SUICIDE;

  board[ pos ] = stone;
  //gethist[ phist ] = count;
  //phist++;

  // ( ( TotalBoard* ) board ) -> capturedStones[ stone - 1 ] += count;

  return 0;
}

int cboard::get( unsigned char *board,
				 int            pos ) {
  short stack[ BALLADE_MAX_BOARD ];
  short p = 0, count = 1;
  stack[ 0 ] = pos;
  unsigned char me = board[ pos ];
  board[ pos ] = (BYTE)NULL;

  mCapturedStone[ pCapturedStone++ ] = pos;

  do {
    short pos0 = stack[ p ];
    for ( int i = 0; i < 4; i++ ) {
      short pos1 = pos0 + mNeighbour[ i ];
      if ( board[ pos1 ] == me ) {
	stack[ p++ ] = pos1;
	mCapturedStone[ pCapturedStone++ ] = pos1;
	board[ pos1 ] = 0;
	count++;
      }
    }
  } while ( --p >= 0 );
  return count;
}

short cboard::getAll( BYTE *board,
		      short pos,
		      int   stone ) {
  BYTE you = BALLADE_ME + BALLADE_YOU - stone;
  short count = 0;

  short pos1 = pos - ( mSize + 1 );
  if ( board[ pos1 ] == you && liberty1( board, pos1 ) == 1 )
	count += get( board, pos1 );
  pos1 = pos - 1;
  if ( board[ pos1 ] == you && liberty1( board, pos1 ) == 1 )
    count += get( board, pos1 );
  pos1 = pos + 1;
  if ( board[ pos1 ] == you && liberty1( board, pos1 ) == 1 )
    count += get( board, pos1 );
  pos1 = pos + ( mSize + 1 );
  if ( board[ pos1 ] == you && liberty1( board, pos1 ) == 1 )
    count += get( board, pos1 );

  return count;
}

int
cboard::ply_first( int stone,
				   uchar *board,
				   Score *pscore,
				   int nmoves,
				   short *moves,
				   unsigned long nPlayout ) {
  double maxScore = -1e99;
  int bestPos = 0;
  for ( int p = 0; p < nmoves; p++ ) {
	int i = moves[ p ];

	if ( pscore[ i ].n > 0 ) {
	  double score = (double) pscore[ i ].sum
		/ (double) ( pscore[ i ].n * mSize * mSize * 2 )
		+ sqrt( log( (double) nPlayout ) / (double) pscore[ i ].n );
	  if ( score > maxScore ) {
		maxScore = score;
		bestPos = i;
	  }
	} else if ( 1 /* i != 0 */ ) {
	  // long code = encode( i, stone, board );
	  // double score = m_pat[ code ];
	  double score = 1e99;
	  if ( score > maxScore ) {
		maxScore = score;
		bestPos = i;
	  }
	}
  }
  return bestPos;

  /*
  while ( true ) {
	int r = rand() % nMoveList;
	if ( r == 0 )
	  return 0;
	else {
	  if ( canPut1( stone, moveList[ r ], board ) )
		return moveList[ r ];
	  else {
		moveList[ r ] = moveList[ -- nMoveList ];
	  }
	}
  }
  */
}

int
cboard::ply( int stone,
			 uchar *board ) {
  int nMoveList = 0;
  short moveList[ BALLADE_MAX_BOARD ];

  for ( int i = 1; i < ( mSize + 1 ) * ( mSize + 1 ); i++ )
	if ( board[ i ] == 0 )
	  moveList[ nMoveList++ ] = i;

  while ( true ) {
	if ( nMoveList == 0 )
	  return 0;
	int r = rand() % nMoveList;
	if ( board[ moveList[ r ] - 1 ] == NULL ||
		 board[ moveList[ r ] + 1 ] == NULL ||
		 board[ moveList[ r ] + ( mSize + 1 ) ] == NULL ||
		 board[ moveList[ r ] - ( mSize + 1 ) ] == NULL )
	  return moveList[ r ];
	else if ( canPut1( stone, moveList[ r ], board ) )
	  return moveList[ r ];
	else
	  moveList[ r ] = moveList[ -- nMoveList ];
  }
}

int
cboard::playout( int p_stone,
				 uchar *board,
				 short *history,
				 Score *pscore,
				 int nmoves,
				 short *moves,
				 unsigned long nPlayout ) {
  int stone = p_stone;
  int prevMove = -1;
  {
	int pos = ply_first( stone, board, pscore, nmoves, moves, nPlayout );
	history[ 0 ] = pos;
	put2( board, stone, pos );
	stone = BALLADE_ME + BALLADE_YOU - stone;
	prevMove = pos;
  }

  for ( int j = 1; j < 1024; j++ ) {
	int pos = ply( stone, board );
	history[ j ] = pos;
	put2( board, stone, pos );
	stone = BALLADE_ME + BALLADE_YOU - stone;
	if ( pos == 0 && prevMove == 0 )
	  break;
	prevMove = pos;
  }

  int stones[ 2 ];
  {
	for ( int i = 0; i < 2; i++ )
	  stones[ i ] = 0; // ( ( TotalBoard* ) board )->capturedStones[ i ];
  }
  for ( int i = 0; i < BALLADE_MAX_BOARD; i++ )
	switch ( board[ i ] ) {
	case BALLADE_ME:
	case BALLADE_YOU:
	  stones[ board[ i ] - 1 ] ++;
	  break;
	case 0:
	  int b = 0, w = 0;
	  switch ( board[ i - mSize - 1 ] ) {
	  case BALLADE_ME: b++; break;
	  case BALLADE_YOU: w++; break;
	  };
	  switch ( board[ i - 1 ] ) {
	  case BALLADE_ME: b++; break;
	  case BALLADE_YOU: w++; break;
	  };
	  switch ( board[ i + 1 ] ) {
	  case BALLADE_ME: b++; break;
	  case BALLADE_YOU: w++; break;
	  };
	  switch ( board[ i + mSize + 1 ] ) {
	  case BALLADE_ME: b++; break;
	  case BALLADE_YOU: w++; break;
	  }
	  if ( b > w )
		stones[ 0 ]++;
	  else if ( b < w )
		stones[ 1 ]++;
	  break;
	}
  // printf( "<%d %d>", stones[ 0 ], stones[ 1 ] );
  return ( p_stone == BALLADE_ME ?
		   stones[ 0 ] - stones[ 1 ] :
		   stones[ 1 ] - stones[ 0 ] );
}

MoveList cboard::mc( void ) {
  Score score[ BALLADE_MAX_BOARD ];
  for ( int i = 0; i < BALLADE_MAX_BOARD; i++ )
	score[ i ].n = score[ i ].sum = 0;

  uchar board[ BALLADE_MAX_BOARD ];

  long startTime = timeGetTime();

  int nMoveList = 1;
  short moveList[ BALLADE_MAX_BOARD ];
  moveList[ 0 ] = 0;

  for ( int i = 1; i < ( mSize + 1 ) * ( mSize + 1 ); i++ )
	if ( mBoard[ i ] == 0 &&
		 ! ( ( mBoard[ i - mSize - 1 ] == BALLADE_YOU || mBoard[ i - mSize - 1 ] == BALLADE_OUTSIDE ) &&
			 ( mBoard[ i         - 1 ] == BALLADE_YOU || mBoard[ i         - 1 ] == BALLADE_OUTSIDE ) &&
			 ( mBoard[ i         + 1 ] == BALLADE_YOU || mBoard[ i         + 1 ] == BALLADE_OUTSIDE ) &&
			 ( mBoard[ i + mSize + 1 ] == BALLADE_YOU || mBoard[ i + mSize + 1 ] == BALLADE_OUTSIDE ) ) )
	  if ( canPut1( BALLADE_ME, i, mBoard ) )
		moveList[ nMoveList++ ] = i;

  for ( int i = 0; true; i++ ) {
	memcpy( board, mBoard, BALLADE_MAX_BOARD );
	short history[ 1024 ];
	int ret = playout( BALLADE_ME, board, history, score, nMoveList, moveList, 99999999 );
	for ( int j = 0; j < 1; j += 2 ) {
	  score[ history[ j ] ].n ++;
	  score[ history[ j ] ].sum += ret;
	}

	long currentTime = timeGetTime();
	if ( currentTime - startTime >= 950 ) {
	  break;
	}
  }

  int bestMove = 0;
  double maxScore = -32768;
  // for ( int i = BOARD_SIZE - 1; i >= 0; i-- )
  {
	for ( int i = 0; i < BALLADE_MAX_BOARD; i++ )
	  if ( score[ i ].n > 0 ) {
		//int s = ( score[ i ].sum + score[ i ].n / 2 )/ score[ i ].n;
		double s = (double) score[ i ].sum / (double)score[ i ].n;
		if ( s > maxScore ) {
		  maxScore = s;
		  bestMove = i;
		}
		// printf( "<%d %d/%d %d %d>", i, s, maxScore, score[ i].sum, score[i].n );
	  }
  }

  MoveList mlist;
  mlist.score	=	(int)maxScore;
  mlist.move	=	bestMove;

  return mlist;
}
