/****************************************************************************
*    This file is part of CEDEC_GO.                                         *
*    Copyright (c) 2004-2011 Takenori Kubo, Youichiro Miyake, CESA          *
*                                                                           *
*    History:                                                               *
*    01/07/2011 Tweaked for CEDEC CHALLANGE 2011 by CESA                    *
*                                                                           *
*    CEDEC_GO is free software: you can redistribute it and/or modify       *
*    it under the terms of the GNU General Public License as published by   *
*    the Free Software Foundation, either version 3 of the License, or      *
*    any later version.                                                     *
*                                                                           *
*    CEDEC_GO is distributed in the hope that it will be useful,            *
*    but WITHOUT ANY WARRANTY; without even the implied warranty of         *
*    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the          *
*    GNU General Public License for more details.                           *
*                                                                           *
*    You should have received a copy of the GNU General Public License      *
*    along with CEDEC_GO.  If not, see <http://www.gnu.org/licenses/>.      *
*****************************************************************************/

#ifndef _ENGINE_H_
#define _ENGINE_H_

#include <string>
 
enum COLOR {
  EMPTY,
  WHITE,
  BLACK,
  BLACK_TERRITORY,
  WHITE_TERRITORY,
  ALIVE,
  ERROR1,
};

enum BOARDSIZE {
  MIN_BOARD = 9,
  MAX_BOARD = 19,
};

class engine {
 public:
  engine(){}
  virtual ~engine() {}

  // for GTP name command
  virtual std::string name() = 0;
  // for GTP version command
  virtual std::string version() = 0;
  // for GTP boardsize, komi, clear_board commands
  virtual void reset(int size, float komi) = 0;
  // for GTP genmove command
  virtual void genmove(int& i, int& j, COLOR color) = 0;
  // for GTP play command
  virtual void playmove(int i, int j, COLOR color) = 0;
  virtual bool legal_move(int i, int j, COLOR color) = 0;

  // for GTP undo command
  virtual void undo() = 0;

  // for GTP showboard command
  virtual COLOR get_board(int i, int j) = 0;

  // for GTP final_score command
  virtual void compute_final_status() = 0;
  virtual COLOR get_final_status(int i, int j) = 0;

  virtual void time_settings(int, int, int) = 0;
  virtual void time_left(COLOR, int, int) = 0;
};

#endif//_ENGINE_H_
